from typing import Optional, Type

from PySide6 import QtWidgets, QtGui, QtCore
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QTextEdit

from sohstationviewer.conf.constants import ROOT_PATH
from sohstationviewer.view.db_config.value_color_helper.\
    edit_value_color_dialog import EditValueColorDialog
from sohstationviewer.view.db_config.value_color_helper.\
    edit_value_color_dialog import LineDotDialog
from sohstationviewer.view.db_config.value_color_helper.\
    edit_value_color_dialog import (
        MultiColorDotLowerEqualDialog, MultiColorDotUpperEqualDialog)
from sohstationviewer.view.db_config.value_color_helper.\
    edit_value_color_dialog import TriColorLinesDialog
from sohstationviewer.view.db_config.value_color_helper.\
    edit_value_color_dialog import UpDownDialog
from sohstationviewer.view.db_config.value_color_helper.\
    edit_value_color_dialog import DotForTimeDialog
from sohstationviewer.view.db_config.value_color_helper.functions import \
    prepare_value_color_html
from sohstationviewer.view.util.plot_type_info import plot_types


plot_types_with_value_colors = [
            p for p in plot_types
            if 'pattern' in plot_types[p]]


class ValueColorEdit(QTextEdit):
    value_color_edited = QtCore.Signal(str)

    def __init__(self, parent: Optional[QWidget], background: str,
                 plot_type: str, value_color_str: str = ''):
        """
        Widget to display valueColors and call a dialog to edit tha value
        :param parent: the parent widget
        :param background: 'B'/'W': flag indicating background color
        :param plot_type: the plot type the editor used to format and validate
            the value colors
        :param value_color_str: the initial value colors shown in the editor.
            If this does not fit the format required by the given plot type,
            use the default value colors for the plot type instead.
        """
        QtWidgets.QTextEdit.__init__(self, parent)
        self.set_background(background)
        # type of channel's plot
        self.plot_type: str = plot_type
        self.value_color_str: str = ''
        if plot_type in plot_types_with_value_colors:
            try:
                self.set_value_color(value_color_str)
            except ValueError:
                # We set the value colors to the default value (specified in
                # plot_type_info.py) when there is a problem with the given
                # value colors string.
                self.set_value_color(
                    plot_types[plot_type]['default_value_color']
                )

        # dialog that pop up when clicking on edit_button to help edit color
        # and value
        self.edit_value_color_dialog: Optional[QWidget] = None

        self.setReadOnly(True)
        # change cursor to Arrow so user know they can't edit directly
        self.viewport().setCursor(
            QtGui.QCursor(QtCore.Qt.CursorShape.ArrowCursor)
        )
        # to see all info
        self.setFixedHeight(28)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.verticalScrollBar().setDisabled(True)

        self.edit_button = QtWidgets.QToolButton(self)
        self.edit_button.setCursor(Qt.PointingHandCursor)
        if background == 'B':
            img_file = f"{ROOT_PATH}/images/edit_icon_black_background.png"
        else:
            img_file = f"{ROOT_PATH}/images/edit_icon_white_background.png"
        self.edit_button.setIcon(QtGui.QIcon(img_file))
        self.edit_button.setStyleSheet(
            "background: transparent; border: none;")
        self.edit_button.clicked.connect(self.edit)
        if plot_type not in plot_types_with_value_colors:
            self.edit_button.hide()

        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(self.edit_button, 0, Qt.AlignRight)

        layout.setSpacing(0)
        layout.setContentsMargins(5, 5, 5, 5)

        self.textChanged.connect(
            lambda: self.value_color_edited.emit(self.value_color_str)
        )

    def set_background(self, background: str):
        """
        Set black background for user to have better feeling how the colors
            displayed on black background. Text and PlaceholderText's colors
            have to be changed to be readable on the black background too.
        :param background: 'B'/'W': sign for background color
        """
        palette = self.palette()
        if background == 'B':
            palette.setColor(QtGui.QPalette.ColorRole.Text, Qt.white)
            palette.setColor(QtGui.QPalette.ColorRole.Base, Qt.black)
            palette.setColor(QtGui.QPalette.ColorRole.PlaceholderText,
                             Qt.lightGray)
            self.setPalette(palette)

    def set_value_color(self, value_color_str: str) -> None:
        """
        Set value_color_str, value_color_edit_dialog and display value color
            string in html to show color for user to have the feeling.
        :param value_color_str: string for value color to be saved in DB
        """
        self.value_color_str = value_color_str
        value_color_html = prepare_value_color_html(self.value_color_str)
        self.setHtml(value_color_html)

    def edit(self):
        """
        Show user an editor to edit the value colors of this widget.
        """
        if self.plot_type not in plot_types_with_value_colors:
            return
        plot_type_dialog_map: dict[str, Type[EditValueColorDialog]] = {
            'linesDots': LineDotDialog,
            'triColorLines': TriColorLinesDialog,
            'dotForTime': DotForTimeDialog,
            'multiColorDotsEqualOnUpperBound': MultiColorDotUpperEqualDialog,
            'multiColorDotsEqualOnLowerBound': MultiColorDotLowerEqualDialog,
            'upDownDots': UpDownDialog,
        }
        edit_dialog = plot_type_dialog_map[self.plot_type](
            self, self.value_color_str
        )
        edit_dialog.accepted.connect(
            lambda: self.set_value_color(edit_dialog.value_color_str)
        )
        edit_dialog.open()
