import numpy as np
from typing import Dict, Tuple, List

from sohstationviewer.conf import constants as const


def get_start_5mins_of_diff_days(start_tm: float, end_tm: float) -> np.ndarray:
    """
    Get the list of the start time of all five minutes for each day start from
        the day of startTm and end at the day of endTm.
    :param start_tm: float - start time
    :param end_tm: float - end time
    :return start_5mins_of_diff_days: [[288 of floats], ] - the list of
        start of all five minutes of days specified by start_tm and end_tm in
        which each day has 288 of 5 minutes.
    """
    exact_day_tm = (start_tm // const.SEC_DAY) * const.SEC_DAY
    exact_day_tm_list = []

    if start_tm < exact_day_tm:
        exact_day_tm_list = [exact_day_tm - const.SEC_DAY]

    while exact_day_tm < end_tm:
        exact_day_tm_list.append(exact_day_tm)
        exact_day_tm += const.SEC_DAY

    # list of start/end 5m in each day: start_5mins_of_diff_days
    for idx, start_day_tm in enumerate(exact_day_tm_list):
        start_5mins_of_day = np.arange(start_day_tm,
                                       start_day_tm + const.SEC_DAY,
                                       const.SEC_5M)
        if idx == 0:
            start_5mins_of_diff_days = np.array([start_5mins_of_day])
        else:
            start_5mins_of_diff_days = np.vstack(
                (start_5mins_of_diff_days, start_5mins_of_day))
    return start_5mins_of_diff_days


def find_tps_tm_idx(
        given_tm: float, start_5mins_of_diff_days: List[List[float]]) \
        -> Tuple[float, float]:
    """
    Find the position of the given time (given_tm) in time-power-squared plot
    :param given_tm: float - given time
    :param start_5mins_of_diff_days: [[288 of floats], ] - the list of
        start of all five minutes of some specific days in which each day has
        288 of 5 minutes.
    :return x_idx: int - index of 5m section
    :return y_idx: int - index of the day the given time belong to in plotting
    """
    x_idx = None
    y_idx = None
    for day_idx, a_day_5mins in enumerate(start_5mins_of_diff_days):
        for start_5m_idx, start_5m in enumerate(a_day_5mins):
            if start_5m > given_tm:
                # index of day start from 0 to negative because day is plotted
                # from top to bottom
                y_idx = - day_idx
                x_idx = start_5m_idx - 1
                if start_5m_idx == 0:
                    # if the start_5m_idx == 0, the given time belong to the
                    # last 5m of the previous day
                    y_idx = -(day_idx - 1)
                    x_idx = const.NO_5M_DAY - 1
                break
        if x_idx is not None:
            break

    if x_idx is None:
        # x_idx == None happens when the given time fall into the last 5m of
        # the last day. Although the time 24:00 of the last day belongs
        # to the next days of other cases, but since there is no more days to
        # plot it, it is no harm to set it at the last 5m of the last day.
        x_idx = const.NO_5M_DAY - 1
        y_idx = - (len(start_5mins_of_diff_days) - 1)

    return x_idx, y_idx


def get_tps_for_discontinuous_data(
        channel_data: Dict,
        start_5mins_of_diff_days: List[List[float]]) -> np.ndarray:
    """
    First loop: look in times for indexes for each block of 5m of each day.
        Because data is discontinuous, some block might have no data points.
    Second loop: For each 5m block, calculate mean of all square of data in
        that block (mean_square). For the blocks that have no data points,
        use the mean of all square of data in the previous and next blocks if
        they both have data or else the mean_square will be zero.

    :param channel_data: dictionary that keeps data of a waveform channel
    :param start_5mins_of_diff_days: the list of starts of all five minutes
        of days in which each day has 288 of 5 minutes.
    :return: array of mean square of five-minute data that are separated into
        days
    """
    times = channel_data['tracesInfo'][0]['times']
    data = channel_data['tracesInfo'][0]['data']
    # create list of indexes for data points of each block of 5m data separated
    # into different days
    tps_idxs = []
    for start5m_of_a_day in start_5mins_of_diff_days:
        tps_idxs.append([])
        for start5m in start5m_of_a_day:
            end5m = start5m + const.SEC_5M
            indexes = np.where((start5m <= times) & (times < end5m))[0]
            tps_idxs[-1].append(indexes)

    # based on tps_idxs, calculated mean square for each 5m data separated into
    # different days
    tps_data = []
    for day_idx in range(len(tps_idxs)):
        tps_data.append([])
        for idx_5m in range(len(tps_idxs[day_idx])):
            try:
                indexes = tps_idxs[day_idx][idx_5m]
                if len(indexes) == 0:
                    # No data point, check both sides, if have data points then
                    # calculate mean square of both sides' data points
                    prev_indexes = tps_idxs[day_idx][idx_5m - 1]
                    if idx_5m < len(tps_idxs[day_idx]) - 1:
                        next_indexes = tps_idxs[day_idx][idx_5m + 1]
                    else:
                        # current 5m block is the last one, the right side
                        # is the first 5m block of the next day
                        next_indexes = tps_idxs[day_idx + 1][0]

                    if len(prev_indexes) != 0 and len(next_indexes) != 0:
                        indexes = np.hstack((prev_indexes, next_indexes))
                if len(indexes) == 0:
                    mean_square = 0
                else:
                    data5m = data[indexes]
                    mean_square = np.mean(np.square(data5m))
            except IndexError:
                mean_square = 0
            tps_data[-1].append(mean_square)

    return np.array(tps_data)


def get_tps_for_continuous_data(channel_data: Dict,
                                start_5mins_of_diff_days: List[List[float]],
                                start_time, end_time):
    """
     Different from soh_data where times and data are each in one np.array,
     in waveform_data, times and data are each kept in a list of np.memmap
     files along with startTmEpoch and endTmEpoch.
     self.channel_data['startIdx'] and self.channel_data['endIdx'] will be
     used to exclude np.memmap files that aren't in the zoom time range
     (startTm, endTm). Data in np.memmap will be trimmed according to times
     then time-power-square value for each 5 minutes will be calculated and
     saved in channel_data['tps-data']: np.mean(np.square(5m data))

     """

    # preset all 0 for all 5 minutes for each day
    tps_data = np.zeros((len(start_5mins_of_diff_days),
                         const.NO_5M_DAY))

    spr = channel_data['samplerate']
    channel_data['tps_data'] = []

    start_tps_tm = 0
    acc_data_list = []

    for tr_idx, tr in enumerate(channel_data['tracesInfo']):
        if 'data_f' in tr:
            times = np.linspace(tr['startTmEpoch'], tr['endTmEpoch'],
                                tr['size'])
            data = np.memmap(tr['data_f'],
                             dtype='int64', mode='r',
                             shape=tr['size'])
        else:
            times = tr['times']
            data = tr['data']
        start_index = 0
        if tr_idx == 0:
            # get index of times with closet value to startTm
            start_index = np.abs(times - start_time).argmin()
            start_tps_tm = times[start_index]

        # identify index in case of overlaps or gaps
        index = np.where(
            (start_5mins_of_diff_days <= times[start_index]) &
            (start_5mins_of_diff_days + const.SEC_5M > times[start_index])
            # noqa: E501
        )
        curr_row = index[0][0]
        curr_col = index[1][0]
        next_tps_tm = start_tps_tm + const.SEC_5M
        while end_time >= next_tps_tm:
            next_index = int(start_index + spr * const.SEC_5M)
            if next_index >= tr['size']:
                acc_data_list.append(data[start_index:tr['size']])
                break
            else:
                acc_data_list.append(
                    np.square(data[start_index:next_index]))
                acc_data = np.hstack(acc_data_list)
                if acc_data.size == 0:
                    tps_data[curr_row, curr_col] = 0
                else:
                    tps_data[curr_row, curr_col] = np.mean(acc_data)

                start_index = next_index
                curr_col += 1
                acc_data_list = []
                if curr_col == const.NO_5M_DAY:
                    curr_col = 0
                    curr_row += 1
                    next_tps_tm += const.SEC_5M
    return tps_data


def get_tps_time_by_color_for_a_day(x: np.ndarray,
                                    y: np.ndarray,
                                    square_counts: List[int],
                                    colors: List[str]) -> Dict[str, List[int]]:
    """
    Get indexes with related color to plot tps
    :param x: array of index of 5 minutes in a day
    :param y: array of index of tps value for related x
    :param square_counts: list of square count levels
    :param colors: color corresponding to square count ranges
    :return tps_x_by_color: dict with color code as key, tps_time which is the
        index of 5 minutes in a day as value to plot tps
    """
    tps_x_by_color = {}
    for i, color in enumerate(colors):
        if i == 0:
            indexes = np.where(y == square_counts[i])[0]
        elif i < len(colors) - 1:
            indexes = np.where((square_counts[i-1] < y) &
                               (y <= square_counts[i]))
        else:
            indexes = np.where(square_counts[i - 1] < y)
        tps_x_by_color[color] = x[indexes]
    return tps_x_by_color
